==> sagua_0.2.0-1.tar.gz: sagua-0.2.0/src/player_wma.cc <==
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// wma decoder came from bmp-wma plugin.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_wma.cc
/// @brief Class for play wma source file.

#include "../include/player_wma.h"
#include "../include/mainmanager.h"

/// @brief Constructor of WMAPlayer.
///
/// Connect dispatchers to functions.
FFMPEG::WMAPlayer::WMAPlayer(MainManager& mainManager)
: decodeState_(STOP),
currentSeconds_(0),
seekSeconds_(0),
bufSize_(4096),
sampleRate_(0),
index_(0),
validFile_(false),
buf_(bufSize_ / 2),
mainManager_(mainManager),
avCodec_(NULL),
avCodecContext_(NULL),
avFormatContext_(NULL),
outbuf_(NULL),
s_outbuf_(NULL)
{
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
    signalTimerUpdate_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayTimeUpdate));
}

FFMPEG::WMAPlayer::~WMAPlayer()
{
    Clean();
}

/// @brief Initialize WMAPlayer.
/// @param filename filename to play
/// @return Success of initialize
bool FFMPEG::WMAPlayer::Init(const std::string& filename)
{
    avcodec_init();
    avcodec_register_all();
    av_register_all();

    if(av_open_input_file(&avFormatContext_, filename.c_str(), NULL, 0, NULL) < 0) 
	return false;
    
    for(index_ = 0; index_ < avFormatContext_->nb_streams; ++index_)
    {
	avCodecContext_ = &avFormatContext_->streams[index_]->codec;
	if(avCodecContext_->codec_type == CODEC_TYPE_AUDIO)
	    break;	
    }

    av_find_stream_info(avFormatContext_);

    avCodec_ = avcodec_find_decoder(avCodecContext_->codec_id);
    if(!avCodec_)
	return false;

    if(avcodec_open(avCodecContext_, avCodec_) < 0) 
	return false;

    outbuf_ = (uint8_t*)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
    s_outbuf_ = (uint8_t*)malloc(bufSize_);

    validFile_ = true;

    return true;
}

/// @brief Start player loop.
///
/// Decode file and send decoed samples to play function. Timer update and
/// audio open, close is also in this function.
void FFMPEG::WMAPlayer::Start()
{
    unsigned int channels = avCodecContext_->channels;
    if(!mainManager_.AudioCheckOpenDevice() || sampleRate_ != avCodecContext_->sample_rate) {
	sampleRate_ = avCodecContext_->sample_rate;
	mainManager_.AudioClose(true);
	mainManager_.AudioOpenDevice(16, channels, sampleRate_, mainManager_.audioFormatNativeEndian_);	
    }

    decodeState_ = PLAY;

    int size = 0;
    int len = 0;
    int out_size = 0;
    uint8_t* inbuf_ptr;

    FifoBuffer f;

    for(;;) {

	{
	    Glib::Mutex::Lock lock(avMutex_);

	    if(decodeState_ == STOP)
		break;
	    else if(decodeState_ == SEEK) {
		av_seek_frame(avFormatContext_, index_, seekSeconds_ * 1000000LL);
		seekSeconds_ = 0;
		decodeState_ = PLAY;
	    }
	}
	
	if(av_read_frame(avFormatContext_, &avPacket_) < 0)
	    break;
	
	size = avPacket_.size;
	inbuf_ptr = avPacket_.data;
	long int seconds = avPacket_.pts/1000000LL;
	if(currentSeconds_ != seconds) {
	    currentSeconds_ = seconds;

	    signalTimerUpdate_();
	}
	
	if(size == 0)
	    break;
	    
	while(size > 0) {
	    len = avcodec_decode_audio(avCodecContext_, 
			(short*)outbuf_,
			&out_size,
			inbuf_ptr,
			size);
	    
	    if(len < 0) 
		break;
	    	
	    if(out_size <= 0)
		continue;	    
	    else {
		fifo_init(&f, out_size*2);
		fifo_write(&f, outbuf_, out_size, &f.wptr);

		const int count = bufSize_ / 4;
		int index = 0;
		while(fifo_read(&f, s_outbuf_, bufSize_, &f.rptr) == 0) {
		    short int* tempSample = reinterpret_cast<short int*>(s_outbuf_);
		    
		    for(int i =0; i < count; ++i) {
			
			index = i * 2;
			buf_[index] = tempSample[i*2];
			buf_[index] += 0.5;
			buf_[index] /= 32767.5;

			index = i * 2 + 1;
			if(channels == 2) {
			    buf_[index] = tempSample[i*2 + 1];
			    buf_[index] += 0.5;
			    buf_[index] /= 32767.5;
			} else
			    buf_[index] = 0.0; 

		    }
		    
		    mainManager_.AudioPlay(buf_, count);
		    
		}

		fifo_free(&f);

	    }
	    
	    size -= len;
	    inbuf_ptr += len;
	    
	    if(avPacket_.data)
		av_free_packet(&avPacket_);	
	}
    }

    if(decodeState_ == PLAY)
	mainManager_.AudioClose(true);
    else
	mainManager_.AudioClose(false);
		
    signalFinish_();
}

/// @brief Set decodeState to STOP.
///
/// Set decodeState to STOP. Detect STOP in Statr() function.
void FFMPEG::WMAPlayer::Stop()
{
    {
	Glib::Mutex::Lock lock(avMutex_);
    
	decodeState_ = STOP;
    }
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds..
///
/// Set decodeState to SEEK. Detect SEEK in Start() function.
/// @param seconds wanted time to seek in seconds.
void FFMPEG::WMAPlayer::Seek(const unsigned long& seconds)
{
    {
	Glib::Mutex::Lock lock(avMutex_);
	
	decodeState_ = SEEK;

	seekSeconds_ = seconds;
    }
}

/// @brief Get current decoded time.
/// @return current decoded time in seconds.
const long int& FFMPEG::WMAPlayer::CurrentTime() const
{
    return currentSeconds_;
}

/// @brief Clean up decoder.
void FFMPEG::WMAPlayer::Clean()
{
    if(s_outbuf_)
	free(s_outbuf_);

    if(outbuf_)
	free(outbuf_);
	
    if(validFile_) {	
	avcodec_close(avCodecContext_);
	av_close_input_file(avFormatContext_);		
    }
}

/// @brief Create WMAPlayer object.
///
/// Used in PlayerFactory.
std::auto_ptr<Player> CreateWMAPlayer(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new FFMPEG::WMAPlayer(mainManager));
}

==> sagua_0.2.1-1.tar.gz: sagua-0.2.1/src/player_wma.cc <==
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// wma decoder came from bmp-wma plugin.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_wma.cc
/// @brief Class for play wma source file.

#include "../include/player_wma.h"
#include "../include/mainmanager.h"

/// @brief Constructor of WMAPlayer.
///
/// Connect dispatchers to functions.
FFMPEG::WMAPlayer::WMAPlayer(MainManager& mainManager)
: decodeState_(STOP),
    currentSeconds_(0),
    seekSeconds_(0),
    bufSize_(4096),
    count_(bufSize_/2),
    sampleRate_(0),
    index_(0),
    validFile_(false),
    buf_(bufSize_),
    mainManager_(mainManager),
    avCodec_(NULL),
    avCodecContext_(NULL),
    avFormatContext_(NULL),
    outbuf_(NULL),
    s_outbuf_(NULL)
{
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
    signalTimerUpdate_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayTimeUpdate));
}

FFMPEG::WMAPlayer::~WMAPlayer()
{
    Clean();
}

/// @brief Initialize WMAPlayer.
/// @param filename filename to play
/// @return Success of initialize
bool FFMPEG::WMAPlayer::Init(const std::string& filename)
{
    avcodec_init();
    avcodec_register_all();
    av_register_all();

    if(av_open_input_file(&avFormatContext_, filename.c_str(), NULL, 0, NULL) < 0) 
	return false;

    for(index_ = 0; index_ < avFormatContext_->nb_streams; ++index_)
    {
	avCodecContext_ = &avFormatContext_->streams[index_]->codec;
	if(avCodecContext_->codec_type == CODEC_TYPE_AUDIO)
	    break;	
    }

    av_find_stream_info(avFormatContext_);

    avCodec_ = avcodec_find_decoder(avCodecContext_->codec_id);
    if(!avCodec_)
	return false;

    if(avcodec_open(avCodecContext_, avCodec_) < 0) 
	return false;

    outbuf_ = (uint8_t*)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
    s_outbuf_ = (uint8_t*)malloc(bufSize_);

    validFile_ = true;

    return true;
}

/// @brief Start player loop.
///
/// Decode file and send decoed samples to play function. Timer update and
/// audio open, close is also in this function.
void FFMPEG::WMAPlayer::Start()
{
    unsigned int channels = avCodecContext_->channels;
    if(!mainManager_.AudioCheckOpenDevice() || sampleRate_ != avCodecContext_->sample_rate) {
	sampleRate_ = avCodecContext_->sample_rate;
	mainManager_.AudioClose(true);
	mainManager_.AudioOpenDevice(16, channels, sampleRate_, mainManager_.audioFormatNativeEndian_);	
    }

    decodeState_ = PLAY;

    int size = 0;
    int len = 0;
    int out_size = 0;
    uint8_t* inbuf_ptr = NULL;
    long int seconds = 0;

    for(;;) {

	{
	    Glib::Mutex::Lock lock(avMutex_);

	    if(decodeState_ == STOP)
		break;
	    else if(decodeState_ == SEEK) {
		av_seek_frame(avFormatContext_, index_, seekSeconds_ * 1000000LL);
		seekSeconds_ = 0;
		decodeState_ = PLAY;
	    }
	}

	if(av_read_frame(avFormatContext_, &avPacket_) < 0)
	    break;

	size = avPacket_.size;
	inbuf_ptr = avPacket_.data;
	seconds = avPacket_.pts/1000000LL;
	if(currentSeconds_ != seconds) {
	    currentSeconds_ = seconds;

	    signalTimerUpdate_();
	}

	if(size == 0)
	    break;

	while(size > 0) {
	    len = avcodec_decode_audio(avCodecContext_, 
		    (short*)outbuf_,
		    &out_size,
		    inbuf_ptr,
		    size);

	    if(len < 0) 
		break;

	    if(out_size <= 0)
		continue;	    
	    else
		Play(out_size, channels);

	    size -= len;
	    inbuf_ptr += len;

	    if(avPacket_.data)
		av_free_packet(&avPacket_);	
	}
    }

    if(decodeState_ == PLAY)
	mainManager_.AudioClose(true);
    else
	mainManager_.AudioClose(false);

    signalFinish_();
}

/// @brief Set decodeState to STOP.
///
/// Set decodeState to STOP. Detect STOP in Statr() function.
void FFMPEG::WMAPlayer::Stop()
{
    Glib::Mutex::Lock lock(avMutex_);

    decodeState_ = STOP;
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds..
///
/// Set decodeState to SEEK. Detect SEEK in Start() function.
/// @param seconds wanted time to seek in seconds.
void FFMPEG::WMAPlayer::Seek(const unsigned long& seconds)
{
    Glib::Mutex::Lock lock(avMutex_);

    decodeState_ = SEEK;

    seekSeconds_ = seconds;
}

/// @brief Get current decoded time.
/// @return current decoded time in seconds.
const long int& FFMPEG::WMAPlayer::CurrentTime() const
{
    return currentSeconds_;
}

/// @brief Clean up decoder.
void FFMPEG::WMAPlayer::Clean()
{
    if(s_outbuf_)
	free(s_outbuf_);

    if(outbuf_)
	free(outbuf_);

    if(validFile_) {	
	avcodec_close(avCodecContext_);
	av_close_input_file(avFormatContext_);		
    }
}

void FFMPEG::WMAPlayer::ProcessStereo(short int* const tempSample, unsigned int& outbufSize, unsigned int& bufIndex, unsigned int& index)
{
    while(outbufSize >= bufSize_) {
	for(unsigned int i = 0; i < count_; ++i) {
	    ShortIntSignalToDoubleSignal(tempSample, bufIndex, index);

	    ShortIntSignalToDoubleSignal(tempSample, bufIndex, index);
	}

	mainManager_.AudioPlay(buf_, count_);
	index = 0;
	outbufSize -= bufSize_;
    }
}

void FFMPEG::WMAPlayer::ProcessMono(short int* const tempSample, unsigned int& outbufSize, unsigned int& bufIndex, unsigned int& index)
{
    while(outbufSize >= bufSize_) {
	for(unsigned int i = 0; i < count_; ++i) {
	    ShortIntSignalToDoubleSignal(tempSample, bufIndex, index);

	    NullSignal(bufIndex, index);
	}
	mainManager_.AudioPlay(buf_, count_);
	index = 0;
	outbufSize -= bufSize_;
    }
}

inline void FFMPEG::WMAPlayer::Play(const int& bufferSize, const unsigned int& channels)
{
    unsigned int index = 0;
    unsigned int bufIndex = 0;
    unsigned int outbufSize = bufferSize / 2;
    short int* tempSample = reinterpret_cast<short int*>(outbuf_);
    if(channels == 2)
	ProcessStereo(tempSample, outbufSize, bufIndex, index);
    else
	ProcessMono(tempSample, outbufSize, bufIndex, index);
}

/// @brief Create WMAPlayer object.
///
/// Used in PlayerFactory.
std::auto_ptr<Player> CreateWMAPlayer(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new FFMPEG::WMAPlayer(mainManager));
}
